#
# Iterated Prisoners Dilemma
# Simulation for GA and Hill Climber
# 60-371 Final Project
#
# Joel Rorseth 104407927
#

#import axelrod as axl
import sys
import random
import matplotlib.pyplot as plt
import numpy as np


# Generic plotting function
def plot(x1, y1, x2, y2, title1, title2, xlab, ylab, title):

    # Plot both lines
    plt.plot(x1, y1, label=title1)
    plt.plot(x2, y2, label=title2)

    # Name the axis
    plt.xlabel(xlab)
    plt.ylabel(ylab)

    # Show metadata
    plt.title(title)
    plt.legend()
    plt.show()

# Generic plotting function
def plot_three(x1, y1, x2, y2, x3, y3, title1, title2, title3, xlab, ylab, title):
    
    # Plot both lines
    plt.plot(x1, y1, label=title1)
    plt.plot(x2, y2, label=title2)
    plt.plot(x3, y3, label=title3)
    
    # Name the axis
    plt.xlabel(xlab)
    plt.ylabel(ylab)
    
    # Show metadata
    plt.title(title)
    plt.legend()
    plt.show()

# Generic plotting function
def plot_single(x1, y1, title1, xlab, ylab, title):
    
    # Plot both lines
    plt.plot(x1, y1, label=title1)
    
    # Name the axis
    plt.xlabel(xlab)
    plt.ylabel(ylab)
    
    # Show metadata
    plt.title(title)
    plt.legend()
    plt.show()





# MARK: Prisoner scoring logic
# Score player's moves against opponent
def player_score(player_moves, opp_moves):
    player_score = 0

    for i in range(len(player_moves)):

        # Adopt this common scoring metric for IPD
        if (player_moves[i] == 'C' and opp_moves[i] == 'C'):
            player_score += 3

        if (player_moves[i] == 'C' and opp_moves[i] == 'D'):
            player_score += 0

        if (player_moves[i] == 'D' and opp_moves[i] == 'C'):
            player_score += 5

        if (player_moves[i] == 'D' and opp_moves[i] == 'D'):
            player_score += 1

    return player_score


# Return score for ym moves against TFT
def my_score_against_tft(player_moves):

    # Define TFT sequence in relation to my moves
    tft_moves = player_moves
    tft_moves.pop()
    tft_moves.insert(0, 'C')

    return player_score(player_moves, tft_moves)


def opponent_score_against_tft(player_moves):
    
    # Define TFT sequence in relation to my moves
    tft_moves = player_moves
    tft_moves.pop()
    tft_moves.insert(0, 'C')
    
    return player_score(tft_moves, player_moves)


# Return score for ym moves against STFT
def my_score_against_stft(player_moves):

    # Define STFT sequence in relation to my moves
    stft_moves = player_moves
    stft_moves.pop()
    stft_moves.insert(0, 'D')

    return player_score(player_moves, stft_moves)


# Return score for ym moves against TF2T
def my_score_against_tf2t(player_moves):

    # Define TFT sequence in reltion to my moves
    tf2t_moves = ['C', 'C']

    for i in range(len(player_moves)-2):
        idx = i+2

        if (player_moves[i] == 'D' and player_moves[i+1] == 'D'):
            tf2t_moves.append('D')
        else:
            tf2t_moves.append('C')

    return player_score(player_moves, tf2t_moves)




#MARK: Genetic Algorithm
# Fitness of a sequence
def seq_fitness(sequence, fit_fn):

    # How does this sequence fair against TFT
    my_score = fit_fn(sequence)
    return my_score


# Select a single parent, with more fit parents being more likely
def selection(pop, pop_scores, pop_fitness):

    rand_fitness = random.uniform(0, 1)
    current_fitness = 0.0
    result = []

    for i in range(len(pop)):

        if current_fitness <= rand_fitness:

            current_fitness += (pop_scores[i] / pop_fitness)
            result = pop[i]

    return result


# Produce a child sequence by crossing two parent sequences
def crossover(p1, p2):

    pos = int(random.random()*len(p1))
    return p1[:pos] + p2[pos:]

    #return (dna1[:pos]+dna2[pos:], dna2[:pos]+dna1[pos:])


# Consider randomly flipping the value at every and any index in sequence
def mutate(child, mut_prob):

    for i in range(len(child)):

        rand_prob = random.uniform(0, 1)

        # If random prob is not larger than small mutation prob, mutate
        if (rand_prob <= mut_prob):

            if (child[i] == 'C'):
                child[i] = 'D'
            else:
                child[i] = 'C'

    return child


# Score of population is score of the best sequence contained within
def score_my_population(pop, fit_fn):

    max_score = -1
    max_seq = []

    for i in range(len(pop)):
        cur_score = fit_fn(pop[i])

        # Update if best seen yet
        if (cur_score > max_score):
            max_score = cur_score
            max_seq = pop[i]

    # Return only score of best seq
    return (max_seq, max_score)


def top_n_in_population(pop, survival_count, fit_fn):

    pop.sort(key=lambda x: seq_fitness(x, fit_fn), reverse=True)
    return pop[:survival_count]


# Return random seq (C's and D's) of length n
def random_seq_of_length(n):

    seq = []
    for _ in range(n):
        seq.append(random.choice(['C','D']))
    return seq


# Generate all successors of a node (mutations in one spot)
def generate_all_successors(node):

    successors = []

    for i in range(len(node)):
        new_node = node

        # Replace current index with flipped bit, now a neighbor
        new_node[i] = 'D' if (new_node[i] == 'C') else 'C'
        successors.append(new_node)

    return successors


# HILL CLIMBING search
def hill_climbing(num_turns, fit_fn):

    # Generate random start node
    current_node = random_seq_of_length(num_turns)
    current_node_fitness = -1
    next_node = []
    next_node_fitness = -1

    while (True):

        # Generate successor to current (all one-off variants)
        successors = generate_all_successors(current_node)

        for neighbor in successors:

            neighbor_fitness = fit_fn(neighbor)

            # If this neighbor beats current, move to it
            if (neighbor_fitness > next_node_fitness):
                next_node_fitness = neighbor_fitness
                next_node = neighbor

        #print("Next", next_node_fitness, ">? Current", fit_fn(current_node))

        # If none of the neighbors improved current, terminate
        if (next_node_fitness <= current_node_fitness):
            return current_node_fitness
            #return (current_node, current_node_fitness)

        #print("Current is now seq with fitness", fit_fn(next_node))
        current_node = next_node
        current_node_fitness = next_node_fitness

    return current_node_fitness
    #return (current_node, current_node_fitness)



# GENETIC ALGORITHM search
def ga(num_gen, pop_size, num_turns, survival_count, mut_prob, fit_fn):


    # Randomly init population
    current_population = [['Z' for x in range(num_turns)] for y in range(pop_size)]
    for p in range(pop_size):
        current_population[p] = random_seq_of_length(num_turns)
        #current_population[p][t] = random.choice(['C','D'])


    #print("> INITIAL POP")
    #for s in current_population:
    #    print(s)
    #print("\n\n")

    # Score each seq in population
    current_pop_scores = []
    for member in current_population:
        current_pop_scores.append( fit_fn(member) )

    total_pop_fitness = sum(current_pop_scores)

    best_seq = []
    best_score = -1

    # For each generation
    for gen in range(num_gen):

        # Init with top 'survival_count' sequences
        next_gen = top_n_in_population(current_population, survival_count, fit_fn)

        #print("> NEW GEN, STARTING WITH TOP FEW")
        #for n in next_gen:
        #    print(n)
        #print()

        # Create new population, 'pop_size' new children
        for i in range(pop_size - survival_count):

            # Select two parents, likely to be more fit parents
            parent1 = selection(current_population, current_pop_scores, total_pop_fitness)
            parent2 = selection(current_population, current_pop_scores, total_pop_fitness)

            # Produce child
            new_child = crossover(parent1, parent2)

            # Randomly mutate the child
            child = mutate(new_child, mut_prob)

            # Add to next generation
            next_gen.append(child)


        # Establish next generation, determine best sequence and its score
        current_population = next_gen
        best_seq, best_score = score_my_population(current_population, fit_fn)

        #print("> FINISHING GEN, POP HAS EVOLVED TO:")
        #for s in current_population:
        #    print(s)
        #print("\n\n")
        #print("Generation", str(gen), ": Best score =", str(best_score))


    #return (best_score, opponent_score_against_tft(best_seq))
    return best_score
    #return (best_seq, best_score)


def test1():

    # TEST 1: Make a range of game lengths to stack up GA vs Hill, graph it?
    # The GA should use TFT? use the best mutation rate and other params

    test_seq_sizes = [5, 10, 20, 30, 40 , 50, 60, 70, 80, 90, 100, 200, 300, 400, 500]
    test_scores_hill = []
    test_scores_ga = []
    
    # For each sequence size we want to compare and graph...
    for size in test_seq_sizes:
        
        hill_scores = []
        ga_scores = []
        
        # Get an average score of GA and Hill for this seq size
        for _ in range(10):
            hill_scores.append(hill_climbing(size, my_score_against_tft))
            ga_scores.append(ga(100, 100, size, 25, 0.015, my_score_against_tft))
        
        # Record avg score for GA and Hill for this seq size
        test_scores_hill.append(np.mean(hill_scores))
        test_scores_ga.append(np.mean(ga_scores))
    
    # Plot results
    plot(test_seq_sizes, test_scores_hill, test_seq_sizes, test_scores_ga,
         "Hill Climbing", "Genetic Algorithm", "Game Length", "Score",
         "TFT: Genetic Algorithm vs. Hill Climbing")


def test3():

    # TEST 3: Compare different mutation rates and graph score for GA only...
    # Vary mut_prob from 0 to 0.3 (aka 0 to 30% ish maybe)

    test_sizes = [10, 20, 50, 80, 100, 150, 200, 250, 300, 400, 500, 1000]
    size_test_scores_ga = []
    for size in test_sizes:
        
        ga_scores = []
        
        # Get an average score of GA and Hill for this seq size
        for _ in range(10):
            ga_scores.append(ga(5, size, 50, 25, 0.015, my_score_against_tft))
        
        # Record avg score for GA and Hill for this seq size
        size_test_scores_ga.append(np.mean(ga_scores))

    # Plot result
    plot_single(test_sizes, size_test_scores_ga,
                "Genetic Algorithm", "Population Size", "Average Score",
                "TFT: Genetic Algorithm Performance with Varied Population Size")


def test4():
    
    # TEST 4
    
    test_seq_sizes = [5, 10, 20, 30, 40 , 50, 60, 70, 80, 90, 100, 200, 300, 400, 500]
    test_scores_opp = []
    test_scores_ga = []
    
    # For each sequence size we want to compare and graph...
    for size in test_seq_sizes:
        
        opp_scores = []
        ga_scores = []
        
        # Get an average score of GA and Hill for this seq size
        for _ in range(10):
            ga_score, op_score = ga(100, 100, size, 25, 0.015, my_score_against_tft)
            ga_scores.append(ga_score)
            opp_scores.append(op_score)
        
        # Record avg score for GA and Hill for this seq size
        test_scores_opp.append(np.mean(opp_scores))
        test_scores_ga.append(np.mean(ga_scores))
    
    # Plot results
    plot(test_seq_sizes, test_scores_opp, test_seq_sizes, test_scores_ga,
         "TFT Opponent", "Genetic Algorithm", "Game Length", "Score",
         "Genetic Algorithm vs. TFT")



# Test a GA on TFT, TF2T and STFT by trying a GA with a range of mutation values
# The performance of three GAs will essentially be compared by common mutation rate
def test_mut_prob(g, p, t, s):
    
    test_probs = [0.0, 0.01, 0.02, 0.05, 0.1, 0.15, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
    test_scores_ga_tft = []
    test_scores_ga_tf2t = []
    test_scores_ga_stft = []
    
    
    for prob in test_probs:
        
        ga_scores_tft = []
        ga_scores_tf2t = []
        ga_scores_stft = []
        
        # Get an average score for GA's over 10 rounds
        for _ in range(10):
            ga_scores_tft.append(ga(g, p, t, s, prob, my_score_against_tft))
            ga_scores_tf2t.append(ga(g, p, t, s, prob, my_score_against_tf2t))
            ga_scores_stft.append(ga(g, p, t, s, prob, my_score_against_stft))
    
        # Record avg score for GA and Hill for this seq size
        test_scores_ga_tft.append(np.mean(ga_scores_tft))
        test_scores_ga_tf2t.append(np.mean(ga_scores_tf2t))
        test_scores_ga_stft.append(np.mean(ga_scores_stft))
        
        print("Finished prob", prob)

    # Plot results
    plot_three(test_probs, test_scores_ga_tft,
        test_probs, test_scores_ga_tf2t,
        test_probs, test_scores_ga_stft,
        "GA vs. TFT", "GA vs. TF2T", "GA vs. STFT",
        "Mutation Probability", "Average Score",
        "Genetic Algorithm Performance with Varied Mutation Probability")



# Test a GA on TFT, TF2T and STFT by trying a GA with a range of num_gens
# The performance of three GAs will essentially be compared by common number of generations
def test_num_gens(p, t, s, m):
    
    test_gens = [5, 10, 20, 40, 50, 70, 100, 150, 200, 250, 300, 500, 600, 700, 900, 1000]
    test_scores_ga_tft = []
    test_scores_ga_tf2t = []
    test_scores_ga_stft = []
    
    
    for gen in test_gens:
        
        ga_scores_tft = []
        ga_scores_tf2t = []
        ga_scores_stft = []
        
        # Get an average score for GA's over 10 rounds
        for _ in range(10):
            ga_scores_tft.append(ga(gen, p, t, s, m, my_score_against_tft))
            ga_scores_tf2t.append(ga(gen, p, t, s, m, my_score_against_tf2t))
            ga_scores_stft.append(ga(gen, p, t, s, m, my_score_against_stft))
    
        # Record avg score for GA and Hill for this seq size
        test_scores_ga_tft.append(np.mean(ga_scores_tft))
        test_scores_ga_tf2t.append(np.mean(ga_scores_tf2t))
        test_scores_ga_stft.append(np.mean(ga_scores_stft))
        
        print("Finished num gens", gen)

    # Plot results
    plot_three(test_gens, test_scores_ga_tft,
        test_gens, test_scores_ga_tf2t,
        test_gens, test_scores_ga_stft,
        "GA vs. TFT", "GA vs. TF2T", "GA vs. STFT",
        "Number of Generations", "Average Score",
        "Genetic Algorithm Performance with Varied Number of Generations")


# Test a GA on TFT, TF2T and STFT by trying a GA with a range of pop sizes
# The performance of three GAs will essentially be compared by common size of population
def test_pop_size(g, t, s_rate, m):
    
    test_pop_sizes = [5, 10, 20, 40, 50, 70, 100, 150, 200, 250, 300, 500, 600, 700, 900, 1000]
    test_scores_ga_tft = []
    test_scores_ga_tf2t = []
    test_scores_ga_stft = []
    
    
    for size in test_pop_sizes:
        
        ga_scores_tft = []
        ga_scores_tf2t = []
        ga_scores_stft = []
        
        # Determine number of survivors for current pop size, given percentage s_rate
        # Eg. 50 in pop, 0.5 s_rate would yield s=25 survivors
        s = int(size * s_rate)
        
        # Get an average score for GA's over 10 rounds
        for _ in range(10):
            ga_scores_tft.append(ga(g, size, t, s, m, my_score_against_tft))
            ga_scores_tf2t.append(ga(g, size, t, s, m, my_score_against_tf2t))
            ga_scores_stft.append(ga(g, size, t, s, m, my_score_against_stft))
        
        # Record avg score for GA and Hill for this seq size
        test_scores_ga_tft.append(np.mean(ga_scores_tft))
        test_scores_ga_tf2t.append(np.mean(ga_scores_tf2t))
        test_scores_ga_stft.append(np.mean(ga_scores_stft))
        
        print("Finished pop size", size)
    
    # Plot results
    plot_three(test_pop_sizes, test_scores_ga_tft,
        test_pop_sizes, test_scores_ga_tf2t,
        test_pop_sizes, test_scores_ga_stft,
        "GA vs. TFT", "GA vs. TF2T", "GA vs. STFT",
        "Population Size", "Average Score",
        "Genetic Algorithm Performance with Varied Population Size")


def test_survivors(g, p, t, m):
    
    test_surv_rates = [0.0, 0.01, 0.02, 0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.45,
                       0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.9, 1.0]
    test_scores_ga_tft = []
    test_scores_ga_tf2t = []
    test_scores_ga_stft = []
    
    
    for rate in test_surv_rates:
        
        ga_scores_tft = []
        ga_scores_tf2t = []
        ga_scores_stft = []
        
        # Determine number of survivors for current pop size, given percentage s_rate
        # Eg. 50 in pop, 0.5 s_rate would yield s=25 survivors
        s = int(p * rate)
        
        # Get an average score for GA's over 10 rounds
        for _ in range(10):
            ga_scores_tft.append(ga(g, p, t, s, m, my_score_against_tft))
            ga_scores_tf2t.append(ga(g, p, t, s, m, my_score_against_tf2t))
            ga_scores_stft.append(ga(g, p, t, s, m, my_score_against_stft))
        
        # Record avg score for GA and Hill for this seq size
        test_scores_ga_tft.append(np.mean(ga_scores_tft))
        test_scores_ga_tf2t.append(np.mean(ga_scores_tf2t))
        test_scores_ga_stft.append(np.mean(ga_scores_stft))
        
        print("Finished survivors =", s)
    
    # Plot results
    plot_three(test_surv_rates, test_scores_ga_tft,
        test_surv_rates, test_scores_ga_tf2t,
        test_surv_rates, test_scores_ga_stft,
        "GA vs. TFT", "GA vs. TF2T", "GA vs. STFT",
        "Survival Rate", "Average Score",
        "Genetic Algorithm Performance with Varied Survival Rate")



def main():

    if (len(sys.argv) != 6):
        print("Usage: python3", sys.argv[0],
            "<num_gen> <pop_size> <num_turns> <mut_prob> <fit_fn>")
        sys.exit()


    num_gen = int(sys.argv[1])
    pop_size = int(sys.argv[2])
    num_turns = int(sys.argv[3])
    mut_prob = float(sys.argv[4])

    fit_fn = my_score_against_tft

    if (sys.argv[5] == "tf2t"):
        fit_fn = my_score_against_tf2t
    if (sys.argv[5] == "stft"):
        fit_fn = my_score_against_stft

    survival_count = int(pop_size / 2)


    #test_mut_prob(100, 100, 50, 25)
    test_num_gens(100, 50, 25, 0.05)
    #test_pop_size(10, 50, 0.25, 0.2)     # Keep 25% of population each generation
    #test_survivors(10, 100, 50, 0.2)

    # TODO: Compare different population sizes... vary pop_size


    # Get scores from respective search algo
    #best_score_hill = hill_climbing(num_turns, fit_fn)
    #print("Hill Climbing final score:", best_score_hill)

    #best_score_ga = ga(num_gen, pop_size, num_turns, survival_count, mut_prob, fit_fn)
    #print("Gen Algorithm final score:", best_score_ga)


main()
